/*
IRC-Transport main.c: Main program loop
Copyright (C) 1999-2000 Keith Minkler <tsbandit@jabber.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You can reach The author at tsbandit@jabber.org (email or jid)
*/

#include <irctrans.h>

typedef struct irc_cmd_parsing_result_st
{
    char *cmd;
    char *from;
    char *next;
} _irc_cmd_parsing_result, *irc_cmd_parsing_result;

int irc_parse_line(char *line, irc_cmd_parsing_result icpr)
{
    icpr->cmd = NULL;
    icpr->from = NULL;
    icpr->next = NULL;

    if(*line == '\0')
	return -1;

    if(*line != ':')
    {
	icpr->cmd = line;
    }
    else
    {  
	icpr->from = line;
	icpr->cmd = strchr(line, ' ');
	if(icpr->cmd == NULL)
	    return -1;
	icpr->cmd[0] = '\0';
	icpr->cmd++;
    }

    icpr->next = strchr(icpr->cmd, ' ');
    if(icpr->next == NULL)
	return -1;

    icpr->next[0] = '\0';
    icpr->next++;

    return 1;
}

void irc_add_to_buffer(irc_session s, char *buffer, int bufsz)
{
    char *temp;

    if(s->buffer == NULL)
    {
        s->buffer = malloc(bufsz + 1);
        s->buffer[0] = '\0';
        strcat(s->buffer, buffer);
        return;
    }

    temp = s->buffer;
    s->buffer = malloc(strlen(temp) + bufsz +1);

    s->buffer[0] = '\0';
    strcat(s->buffer, temp);
    strcat(s->buffer, buffer);

    free(temp);
}

char *irc_get_next_line(irc_session s, pool p)
{
    char *ptr;
    char *ret;

    if(s->buffer == NULL)
        return NULL;

    ptr = strchr(s->buffer, '\r');

    if(ptr == NULL)
        ptr = strchr(s->buffer, '\n');

    if(ptr == NULL)
        return NULL;

    ptr[0] = '\0';
    ptr++;

    while(ptr[0] == '\r' || ptr[0] == '\n') ptr++;

    ret = pstrdup(p, s->buffer);

    if(ptr[0] != '\0')
    {
        char *temp = s->buffer;
        s->buffer = malloc(strlen(ptr) + 1);
        s->buffer[0] = '\0';
        strcat(s->buffer, ptr);
        free(temp); 
    }
    else
    {
        free(s->buffer);
        s->buffer = NULL;
    }

    return ret;
}

void __irc_cleanup_channel_users(xht users, const char *key, void *val, void *arg)
{
    user u = (user) val;
    log_debug(ZONE, "[__irc_cleanup_channel_users] ZAPPING user %s key %s id %s", u->user, key, jid_full(u->id));

    /* zap the two entries */
    if (strcmp(key,jid_full(u->id)) == 0) {
	xhash_zap(users, key);
	xhash_zap(users, u->user);
    } else {
	xhash_zap(users, key);
	xhash_zap(users, jid_full(u->id));
    }

    pool_free(u->p);
}

void _irc_cleanup_channel(xht channels, const char *key, void *val, void *arg)
{
    chans channel = (chans) val;
    irc_session s = (irc_session) channel->data;

    if (s == NULL) {
	log_error(ZONE,"[_irc_cleanup_channel] unable to retrieve session struct !");
	return;
    }

    if (strcmp(key,channel->id) == 0) {
	xhash_zap(channels, key);
	xhash_zap(channels, channel->name);
    } else {
	xhash_zap(channels, key);
	xhash_zap(channels, channel->id);
    }

    log_debug(ZONE, "[_irc_cleanup_channel] shutting down channel %s on server %s for user %s", channel->name, s->server, jid_full(s->j));
    xhash_walk(channel->users, __irc_cleanup_channel_users, arg);
    xhash_free(channel->users);
    pool_free(channel->p);
}

void _irc_check_quit(xht channels, const char *key, void *val, void *arg)
{
    chans channel = (chans)val;
    irc_session s;
    user u = xhash_get(channel->users, arg);

    if(u)
    {
        jpacket jp = jpacket_new(xmlnode_new_tag("iq"));
        xmlnode x;

        channel = (chans)u->data;
        s = (irc_session)channel->data;

	log_debug(ZONE,"[_irc_check_quit] sending browse remove to %s", jid_full(s->j));

        xmlnode_put_attrib(jp->x, "to", jid_full(s->j));
        xmlnode_put_attrib(jp->x, "from", channel->id);
        xmlnode_put_attrib(jp->x, "type", "set");
	x = xmlnode_insert_tag(jp->x, "user");
	xmlnode_put_attrib(x, "xmlns", NS_BROWSE);
	xmlnode_put_attrib(x, "jid", jid_full(u->id));
        xmlnode_put_attrib(x, "type", "remove");

        deliver(dpacket_new(jp->x), NULL);

	/* send departure notice */
	if (ircdata->note_leave != NULL) {
	    xmlnode msg,body;

	    msg = xmlnode_new_tag("message");
	    xmlnode_put_attrib(msg,"to",jid_full(s->j));
	    xmlnode_put_attrib(msg,"type","groupchat");
	    xmlnode_put_attrib(msg,"from",channel->id);
	    body = xmlnode_insert_tag(msg,"body");
	    xmlnode_insert_cdata(body, u->user, -1);
	    xmlnode_insert_cdata(body, " ", -1);
	    xmlnode_insert_cdata(body, ircdata->note_leave, -1);
	    deliver(dpacket_new(msg), NULL);
	}

        xhash_zap(channel->users, u->user);
        xhash_zap(channel->users, jid_full(u->id));
        pool_free(u->p);
    }
}

typedef struct irc_nick_change_st
{
    char *old_nick;
    char *new_nick;
} _irc_nick_change, *irc_nick_change;

void _update_nick(xht channels, const char *key, void *val, void *arg)
{
    chans channel = (chans) val;
    irc_session s;
    irc_nick_change nc = (irc_nick_change) arg;

    /* check to see if the user is in this channel */
    user u = xhash_get(channel->users, nc->old_nick);
    if (u)
    {
        /* user found, update it */

        jpacket jp = jpacket_new(xmlnode_new_tag("iq"));
        s = (irc_session)channel->data;

        /* create a browse update for the new nick */
        xmlnode_put_attrib(jp->x, "to", jid_full(s->j));
        xmlnode_put_attrib(jp->x, "from", channel->id);
        xmlnode_put_attrib(jp->x, "type", "set");
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x, "user"), "xmlns", NS_BROWSE);
        jpacket_reset(jp);

        xmlnode_put_attrib(jp->iq, "jid", jid_full(u->id));
        xmlnode_put_attrib(jp->iq, "name", nc->new_nick);
        deliver(dpacket_new(jp->x), NULL);

	/* send rename notice */
	if (ircdata->note_rename != NULL) {
	    xmlnode msg,body;

	    msg = xmlnode_new_tag("message");
	    xmlnode_put_attrib(msg,"to",jid_full(s->j));
	    xmlnode_put_attrib(msg,"type","groupchat");
	    xmlnode_put_attrib(msg,"from",channel->id);
	    body = xmlnode_insert_tag(msg,"body");
	    xmlnode_insert_cdata(body, u->user, -1);
	    xmlnode_insert_cdata(body, " ", -1);
	    xmlnode_insert_cdata(body, ircdata->note_rename, -1);
	    xmlnode_insert_cdata(body, " ", -1);
	    xmlnode_insert_cdata(body, nc->new_nick, -1);
	    deliver(dpacket_new(msg), NULL);
	}

        /* change the hash key for the username */
        xhash_zap(channel->users, u->user);
        u->user = make_lowercase(pstrdup(u->p, nc->new_nick));
	log_debug(ZONE,"[_update_nick] updating %s in channel users table", u->user);
        xhash_put(channel->users, u->user, (void *) u);
    }
}

/* phase two handles everything after the initial join.. there will
 * always be a valid session for phase two */
void irc_parse_phase_two(mio m, int flag, void *arg, char *buffer, int bufsz)
{
    irc_session s = (irc_session)arg;
    char *line;
    pool p;

    log_debug(ZONE, "[irc_parse_phase_two] in phase two: m%X flag:%d, arg:%X, buffer: %s, bufsz:%d", m, flag, arg, buffer, bufsz);

    if(flag == MIO_ERROR) { 
        /* got an error, clean out the mio queue */
        xmlnode x;
	log_error(ZONE,"[irc_parse_phase_two] Lost socket connection for %s", jid_full(s->j));
        while ((x = mio_cleanup(m)) != NULL)
            deliver_fail(dpacket_new(x), "IRC Connection Failed");
        return;
    }

    if(flag == MIO_CLOSED) { 
        /* connection to IRC server closed */
        xhash_zap(ircdata->sessions, s->id);
        xhash_walk(s->channels, _irc_cleanup_channel, NULL);
        xhash_free(s->channels);
        if(s->buffer != NULL)
            free(s->buffer);
        pool_free(s->p);
        return;
    }

    /* set this MIO object to the session */
    s->m = m;

    /* add the irc server buffer to our buffer */
    irc_add_to_buffer(s, buffer, bufsz);

    /* temporary pool */
    p = pool_heap(1024);

    while((line = irc_get_next_line(s, p)) != NULL)
    { 
        /* look at commands one line at a time */
        irc_cmd_parsing_result icpr = (irc_cmd_parsing_result) malloc(sizeof(_irc_cmd_parsing_result));

	if (irc_parse_line(line,icpr) == -1)
	    continue;

        /* check if we are interested in any of these commands */
        if(j_strcmp(icpr->cmd, "PING") == 0)
        {
            mio_write(m, NULL, spools(p, "PONG ", icpr->next, "\n", p), -1);
        }
        if(j_strcmp(icpr->cmd, "JOIN") == 0)
        {
            user u;
            pool p2;
            chans channel;
            jpacket jp;
            xmlnode x;
            char *ptr;

            ++icpr->next;
            channel = xhash_get(s->channels, make_lowercase(icpr->next));
            if(icpr->from == NULL)
                continue;
	    if (channel == NULL) {
		log_error(ZONE,"[irc_parse_phase_two] Unable to get channel %s", icpr->next);
	    }
            ptr = strchr(++icpr->from, '!');
            ptr[0] = '\0';

	    /* someone is joining the channel */
            p2 = pool_new();
            u = pmalloco(p2, sizeof(_user));
            u->p = p2;
            u->user = make_lowercase(pstrdup(u->p, icpr->from));
            u->id = jid_new(u->p, spools(p, channel->id, "/", shahash(u->user), p));
            u->data = (void*)channel;
	    log_debug(ZONE,"[irc_parse_phase_two] Adding %s and %s in channel users table", jid_full(u->id), u->user);
            xhash_put(channel->users, jid_full(u->id), u);
            xhash_put(channel->users, u->user, u);

	    /* send the iq browse set packet */
            jp = jpacket_new(xmlnode_new_tag("iq"));
	    x = xmlnode_insert_tag(jp->x, "user");
	    xmlnode_put_attrib(x, "xmlns", NS_BROWSE);
	    xmlnode_put_attrib(x, "jid", jid_full(u->id));
            xmlnode_put_attrib(x, "name", u->user);
            xmlnode_put_attrib(jp->x, "to", jid_full(s->j));
            xmlnode_put_attrib(jp->x, "from", channel->id);
            xmlnode_put_attrib(jp->x, "type", "set");
            deliver(dpacket_new(jp->x), NULL);

	    /* and the join notice message */
	    if (ircdata->note_join != NULL) {
		xmlnode body;
		x = xmlnode_new_tag("message");
		xmlnode_put_attrib(x,"to", jid_full(s->j));
		xmlnode_put_attrib(x,"from", channel->id);
		xmlnode_put_attrib(x,"type", "groupchat");
		body = xmlnode_insert_tag(x,"body");
		xmlnode_insert_cdata(body, u->user, -1);
		xmlnode_insert_cdata(body, " ", -1);
		xmlnode_insert_cdata(body, ircdata->note_join, -1);
		deliver(dpacket_new(x), NULL);
	    }
        }
        else if(j_strcmp(icpr->cmd, "NICK") == 0)
        {
            char *ptr;
            irc_nick_change nc;

            ptr = strchr(++icpr->from, '!');
            if(ptr == NULL) continue;

            ptr[0] = '\0';

	    log_debug(ZONE,"[irc_parse_phase_two] nickname change to %s from %s", icpr->from, s->nick);

            if(j_strcmp(icpr->from,s->nick) == 0)
            {
                /* this is our nick change, we must have a jpacket on s->aux1 */
                jpacket jp = s->aux1;

                log_debug(ZONE,"[irc_parse_phase_two] got a response to our nick change !");
		s->aux1 = NULL;
		/* update our nickname */
		/* added by BO, 2001/10/16 */
		icpr->next++;
		s->nick = pstrdup(s->p, icpr->next);
		icpr->next--;
                if(jp == NULL)
                    continue; /* something has gone wrong */
                jutil_iqresult(jp->x);
                deliver(dpacket_new(jp->x), NULL);
            }

            nc = pmalloco(p, sizeof(_irc_nick_change));
            nc->old_nick = make_lowercase(icpr->from);
            nc->new_nick = make_lowercase(++icpr->next);
            xhash_walk(s->channels, _update_nick, (void*)nc);
        }
        else if(j_strcmp(icpr->cmd, "433") == 0) /* NICK IN USE */
        {
            jpacket jp = s->aux1;

            if (jp == NULL)
                continue;

            if (xmlnode_get_tag_data(jp->iq, "nickname") == NULL)
            {
                /* no more nicks to try */
                jutil_error(jp->x, TERROR_CONFLICT);
                deliver(dpacket_new(jp->x), NULL);
                s->aux1 = NULL;
                return;
            }

            mio_write(m, NULL, spools(p, "NICK :", xmlnode_get_tag_data(jp->iq, "nick"), "\n", p), -1);
            xmlnode_hide(xmlnode_get_tag(jp->iq, "nick"));
        }
	/* XXX : send messages to the client to inform it of chgt */
	/* by BO, 2001/10/02 */
        else if(j_strcmp(icpr->cmd, "MODE") == 0)
	{
	  log_debug(ZONE,"[irc_parse_phase_two] MODE not implemented !");
        }
        else if(j_strcmp(icpr->cmd, "PRIVMSG") == 0)
        {
            /* message from the wonderful world of IRC */
            chans channel;
            xmlnode x;
            char *ptr,*msg;
            user u;

            ptr = strchr(icpr->next, ' ');
            if(ptr == NULL) continue;
            ptr[0] = '\0';
            msg = ptr+1;

            ptr = strchr(++icpr->from, '!');
            if(ptr == NULL) continue;
            ptr[0] = '\0';

            channel = xhash_get(s->channels, make_lowercase(icpr->next));

	    /* prepare the message packet */
            x = xmlnode_new_tag("message");
            xmlnode_put_attrib(x, "to", jid_full(s->j));

	    /* private message : send it as a normal jabber msg */
	    log_debug(ZONE,"[irc_parse_phase_two] PRIVMSG : comparing %s and %s", icpr->next, s->nick);
	    if (j_strcasecmp(icpr->next, s->nick) == 0) {
		xmlnode_put_attrib(x, "type", "normal");
		xmlnode_put_attrib(x, "from", icpr->from);
	    } else {
		if(channel == NULL) {
		    log_warn(ZONE, "[irc_parse_phase_two] channel not found for private message to %s", icpr->next);
		    continue;
		}
	    
		u = xhash_get(channel->users, make_lowercase(icpr->from)); 
		if(u == NULL) {
		    log_debug(ZONE,"[irc_parse_phase_two] user %s is not in the channel", icpr->from);
		    continue;
		}

		xmlnode_put_attrib(x, "from", jid_full(u->id));
		xmlnode_put_attrib(x, "type", "groupchat");
	    }

            icpr->next = ++msg;
            if(*icpr->next == '\001')
            {
                char *ptr;
                icpr->next++;
                if(j_strncasecmp(icpr->next, "ACTION", 6) == 0)
                {
                    ptr = strchr(icpr->next, '\001');
                    ptr[0] = '\0';
                    icpr->next +=3;
                    icpr->next[0] = '/';
                    icpr->next[1] = 'm';
                    icpr->next[2] = 'e';
                }
                else
                {
                    xmlnode_free(x);
                    continue;
                }
            }
            irc_strip_mirc_codes(icpr->next);
            icpr->next = str_to_UTF8(xmlnode_pool(x), icpr->next);
            xmlnode_insert_cdata(xmlnode_insert_tag(x, "body"), icpr->next, -1);

            deliver(dpacket_new(x), NULL);
        }
	/* XXX : retrieve the quit message if it exists */
        else if(j_strcmp(icpr->cmd, "QUIT") == 0)
        {
            char *ptr;

            ptr = strchr(++icpr->from, '!');
            if(ptr == NULL) continue;

            ptr[0] = '\0';

            icpr->from = make_lowercase(icpr->from);

            xhash_walk(s->channels, _irc_check_quit, (void*)icpr->from);
        }
	/* XXX : retrieve the part message if it exists */
        else if(j_strcmp(icpr->cmd, "PART") == 0) 
        {
            user u;
            chans channel;
            jpacket jp;
            xmlnode x;
            char *ptr;

            ptr = strchr(icpr->next, ' ');
            if(ptr != NULL)
                ptr[0] = '\0';

            channel = xhash_get(s->channels, make_lowercase(icpr->next));
            if(icpr->from == NULL || channel == NULL)
                continue;

            ptr = strchr(++icpr->from, '!');
            ptr[0] = '\0';

            u = xhash_get(channel->users, make_lowercase(icpr->from));
            if(u == NULL)
                continue;

            xhash_zap(channel->users, jid_full(u->id));
            xhash_zap(channel->users, u->user);

            jp = jpacket_new(xmlnode_new_tag("iq"));
            xmlnode_put_attrib(jp->x, "to", jid_full(s->j));
            xmlnode_put_attrib(jp->x, "from", channel->id);
            xmlnode_put_attrib(jp->x, "type", "set");
	    x = xmlnode_insert_tag(jp->x, "user");
            xmlnode_put_attrib(x, "jid", jid_full(u->id));
            xmlnode_put_attrib(x, "type", "remove");
	    xmlnode_put_attrib(x, "xmlns", NS_BROWSE);
            deliver(dpacket_new(jp->x), NULL);

	    /* send leave notice */
	    if (ircdata->note_leave != NULL) {
		xmlnode msg,body;

		msg = xmlnode_new_tag("message");
		xmlnode_put_attrib(msg,"to",jid_full(s->j));
		xmlnode_put_attrib(msg,"type","groupchat");
		xmlnode_put_attrib(msg,"from",channel->id);
		body = xmlnode_insert_tag(msg,"body");
		xmlnode_insert_cdata(body, u->user, -1);
		xmlnode_insert_cdata(body, " ", -1);
		xmlnode_insert_cdata(body, ircdata->note_leave, -1);
		deliver(dpacket_new(msg), NULL);
	    }
        } else if (j_strcmp(icpr->cmd, "TOPIC") == 0) 
	  {
	    char *ptr,*topic;
	    xmlnode msg,subject,body;
	    chans channel;

	    log_debug(ZONE,"[irc_parse_phase_two] received request to change topic !");
	    ptr = strchr(icpr->next,' ');
	    if (ptr == NULL)
	      continue;
	    ptr[0] = '\0';
	    ptr++;
	    topic = ptr;
	    
	    channel = xhash_get(s->channels, make_lowercase(icpr->next));
	    if (channel == NULL) {
	      log_error(ZONE,"[irc_parse_phase_two] unable to retrieve channel");
	      continue;
	    }
	    channel->topic = pstrdup(channel->p,topic);
	    /* send topic change notice */
	    msg = xmlnode_new_tag("message");
	    xmlnode_put_attrib(msg, "from", channel->id);
            xmlnode_put_attrib(msg, "to", jid_full(s->j));
	    xmlnode_put_attrib(msg, "type", "groupchat");
	    subject = xmlnode_insert_tag(msg,"subject");
	    xmlnode_insert_cdata(subject,topic,-1);
	    body = xmlnode_insert_tag(msg,"body");
	    if (icpr->from != NULL)
		xmlnode_insert_cdata(body,icpr->from,-1);
	    xmlnode_insert_cdata(body, " ", -1);
	    xmlnode_insert_cdata(body, ircdata->note_topic, -1);
	    xmlnode_insert_cdata(body, topic, -1);
	    deliver(dpacket_new(msg),NULL);
	  }
    }

    pool_free(p);
}

/* phase one will handle the initial channel join */
void irc_parse_phase_one(mio m, int flag, void *arg, char *buffer, int bufsz)
{ 
    /* second phase of creating a session */
    /* we still have the joining jpacket laying around */
    jpacket jp = (jpacket)arg; 
    irc_session s = (irc_session)jp->aux1;
    char *line;
    pool p;

    log_debug(ZONE, "[irc_parse_phase_one] in phase one: m%X flag:%d, arg:%X, buffer: %s, bufsz:%d", m, flag, arg, buffer, bufsz);

    if(flag == MIO_ERROR) { 
        /* XXX umm... shouldn't this be a bit more robust? */
        return;
    }

    if(flag == MIO_CLOSED) { 
        /* we shouldn't be disconnected here yet.. */
        xmlnode_put_attrib(jp->x, "type", "error");
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "error"), "Socket Error to IRC Server", -1);
        xmlnode_put_attrib(xmlnode_get_tag(jp->x, "error"), "code", "502");
        deliver(dpacket_new(jp->x), NULL);

        /* cleanup the session object */
        /* XXX should this only be cleaned up when the user sends unavailabe presence? */
        xhash_zap(ircdata->sessions, spools(s->p, jid_full(s->j), "/", s->server, s->p));
        xhash_walk(s->channels, _irc_cleanup_channel, NULL);
        xhash_free(s->channels);
        if(s->buffer != NULL)
            free(s->buffer);
        pool_free(s->p);
        return;
    }

    /* add the irc server text to the buffer */
    irc_add_to_buffer(s, buffer, bufsz);

    /* create a working pool */
    p = pool_new();

    while((line = irc_get_next_line(s, p)) != NULL)
    { 
        /* read each line of text from the server at a time */
        irc_cmd_parsing_result icpr = (irc_cmd_parsing_result) malloc(sizeof(_irc_cmd_parsing_result));

	if (irc_parse_line(line,icpr) == -1)
	    continue;

        /* look for useful commands */
        if(j_strcmp(icpr->cmd, "PING") == 0)
        {
            mio_write(m, NULL, spools(p, "PONG ", icpr->next, "\n", p), -1);
        }
        else if(j_strcmp(icpr->cmd, "JOIN") == 0)
        { 
            /* we have joined a channel */
            jpacket new;

            /* send out iq:result, and work on iq browse packet */
            jutil_iqresult(jp->x);
	    if (s->state == sess_CREATE) {
	        xmlnode enter;
		pool p = pool_new();
		xmlnode_put_attrib(xmlnode_insert_tag(jp->x, "create"), "xmlns", NS_CONFERENCE);
		jpacket_reset(jp);
		/* insert address and nickname */
		enter = xmlnode_insert_tag(jp->iq, "enter");
		xmlnode_insert_cdata(xmlnode_insert_tag(enter, "nickname"), s->nick, -1);
		/* get channel name */
		icpr->next++;
		xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "address"), spools(p, icpr->next, "@", jid_full(jp->from), p), -1);
		pool_free(p);
	    } else {
	        xmlnode_put_attrib(xmlnode_insert_tag(jp->x, "enter"), "xmlns", NS_CONFERENCE);
		jpacket_reset(jp);
		xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "nickname"), s->nick, -1);
	    }

            /* create the new iq:browse jpacket */
            new = jpacket_new(xmlnode_new_tag("iq"));
            xmlnode_put_attrib(new->x, "type", "set");
	    /* set the correct "from" depending on session type */
	    if (s->state == sess_ENTER) {
	      xmlnode_put_attrib(new->x, "from", jid_full(jp->from));
	    } else if (s->state == sess_CREATE) {
	      pool p = pool_new();
	      xmlnode_put_attrib(new->x, "from", spools(p, icpr->next, "@", jid_full(jp->from), p));
	      pool_free(p);
	    }
            xmlnode_put_attrib(new->x, "to", jid_full(s->j));
            xmlnode_put_attrib(xmlnode_insert_tag(new->x, "conference"), "xmlns", NS_BROWSE);
            jpacket_reset(new);
            xmlnode_put_attrib(new->iq, "type", "irc");
            xmlnode_put_attrib(new->iq, "jid", jid_full(jp->from));

            /* deliver the iq:conference jpacket */
            deliver(dpacket_new(jp->x), NULL);

            /* reset the current jpacket to be the browse packet */
            new->aux1 = s;
            jp = new;
            mio_reset(m, irc_parse_phase_one, (void*)new);
        }
	/* RPL_TOPIC -- <channel> :<topic> */
	else if (j_strcmp(icpr->cmd, "332") == 0)
	{
	    char *ptr,*topic;
	    xmlnode msg,subject;

	    ptr = strchr(icpr->next,':');
	    ptr[0] = '\0';
	    ptr++;
	    topic = ptr;

	    /* send topic */
	    msg = xmlnode_new_tag("message");
	    xmlnode_put_attrib(msg, "from", jid_full(jp->from));
            xmlnode_put_attrib(msg, "to", jid_full(s->j));
	    xmlnode_put_attrib(msg, "type", "groupchat");
	    subject = xmlnode_insert_tag(msg,"subject");
	    xmlnode_insert_cdata(subject,topic,-1);
	    deliver(dpacket_new(msg),NULL);
	}
	/* RPL_NAMREPLY -- can receive multiple 353 if large roster */
        else if (j_strcmp(icpr->cmd, "353") == 0)
        { 
            /* :me 353 mynick = #channel :@me +you them */
            char *ptr;
            chans channel;
            
            ptr = strchr(icpr->next, '=');
            if(ptr == NULL)
                continue;
            icpr->next = ++ptr;
            ptr = ++icpr->next;
            icpr->next = strchr(ptr, ' ');
            if(icpr->next == NULL)
                continue;
            icpr->next[0] = '\0';

            /* find this channel in our xhash of channels */
            channel = xhash_get(s->channels, make_lowercase(ptr));
            if(channel == NULL) { 
	        /* this channel doesn't exist.. create it */
                pool p_chan = pool_new();
                channel = pmalloco(p_chan, sizeof(_chans));
                channel->p = p_chan;
                channel->state = chan_JOINING;
		channel->name = irc_get_channel(p_chan, jid_full(jp->from));
		channel->id = pstrdup(p_chan, jid_full(jp->from));
                channel->users = xhash_new(101);
                channel->data = (void*)s;
                channel->nick_id = spools(p_chan, channel->id, "/", shahash(s->nick), p_chan);

                /* put this channel into the channel hash */
		log_debug(ZONE,"[irc_parse_phase_one] storing channel names : %s and %s", channel->id, channel->name);
                xhash_put(s->channels, channel->id, channel);
                xhash_put(s->channels, channel->name, channel);
            }

            ptr = ++icpr->next;
            ptr++;

            /* get the list of users */
            while(ptr != NULL)
            {
                pool p_user;
                user u;
                xmlnode x;
                icpr->next = strchr(ptr, ' ');
                if(icpr->next != NULL)
                {
                    icpr->next[0] = '\0';
                    icpr->next++;
                }

                p_user = pool_new();
                u = pmalloco(p_user, sizeof(_user));
                u->p = p_user;
                if(*ptr == '@')
                {
                    u->ops = 1;
                    ptr++;
                }
                else if(*ptr == '+')
                {
                    u->voice = 1;
                    ptr++;
                }
                u->user = make_lowercase(pstrdup(u->p, ptr));
		u->id   = jid_new(u->p, spools(u->p, jid_full(jp->from), "/", shahash(u->user), u->p));
                u->data = (void*)channel;

                ptr = icpr->next;
                if(ptr != NULL && *ptr == '\0')
                    ptr = NULL;

                /* add this user to the channel's user hash */
		log_debug(ZONE,"[irc_parse_phase_one] storing user : %s and %s", jid_full(u->id), u->user);
                xhash_put(channel->users, jid_full(u->id), u);
                xhash_put(channel->users, u->user, u);

                /* put this user into the iq:browse packet */
                x = xmlnode_insert_tag(jp->iq, "user");
                xmlnode_put_attrib(x, "jid", jid_full(u->id));
                xmlnode_put_attrib(x, "name", u->user);
            }
        }
	/* end of /NAMES list */
        else if(j_strcmp(icpr->cmd, "366") == 0) 
        { 
	    /* no more users in the channel.. send the iq:browse packet */
            deliver(dpacket_new(jp->x), NULL);

            /* kickstart phase two */
            mio_reset(m, irc_parse_phase_two, (void*)s);
            mio_write(m, NULL, "PING :foo\n", -1);
        }
    }

    pool_free(p);
}


/* in this phase, we are logging into the server,
 * setting our nick, and
 * sending the initial JOIN request */
void irc_parse_phase_zero(mio m, int flag, void *arg, char *buffer, int bufsz)
{
    jpacket jp = (jpacket)arg;
    irc_session s = (irc_session)jp->aux1;
    char *line;
    pool p;

    log_debug(ZONE, "[irc_parse_phase_zero] in phase zero: m%X flag:%d, arg:%X, buffer: %s, bufsz:%d", m, flag, arg, buffer, bufsz);

    if(flag == MIO_ERROR)
        return;

    /* socket to IRC server is established 
     * can now log into the server ... */
    if(flag == MIO_NEW)
    {
        char *nick;

	nick = xmlnode_get_tag_data(jp->iq, "nickname");
        if(nick == NULL)
            nick = jp->from->user;

	/* IDX IRC SPECIAL CASE */
#ifdef IDX_IRC
	/* send the password before USER/NICK commands */
	mio_write(m, NULL, spools(s->p, "PASS ", IDX_IRC_PASSWD, "\n", s->p), -1);
#endif

	/* copy nick in the session struct */
	s->nick = pstrdup(s->p,nick);
        xmlnode_hide(xmlnode_get_tag(jp->iq, "nickname"));

        mio_write(m, NULL, spools(s->p, "USER ", nick, " 0 * :Jay-Jay Okocha\n", s->p), -1);
        mio_write(m, NULL, spools(s->p, "NICK ", nick, "\n",  s->p), -1);
        return;
    }

    if(flag == MIO_CLOSED)
    {
	if (jp->aux1 != NULL) {
	    xmlnode_put_attrib(jp->x, "type", "error");
	    xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "error"), "Socket Error to IRC Server", -1);
	    xmlnode_put_attrib(xmlnode_get_tag(jp->x, "error"), "code", "502");
	}
	deliver(dpacket_new(jp->x), NULL);
        return;
    }

    /* add the text from irc server into the buffer */
    irc_add_to_buffer(s, buffer, bufsz);

    /* create the working pool */
    p = pool_new();

    while((line = irc_get_next_line(s, p)) != NULL)
    { 
        /* for each line of input... */
        irc_cmd_parsing_result icpr = (irc_cmd_parsing_result) malloc(sizeof(_irc_cmd_parsing_result));

	if (irc_parse_line(line,icpr) == -1)
	    continue;

        /* see if this is a command we are interested in */
        if(j_strcmp(icpr->cmd, "PING") == 0)
        {
            mio_write(m, NULL, spools(p, "PONG ", icpr->next, "\n", p), -1);
            continue;
        }
        else if(j_strcmp(icpr->cmd, "433") == 0) /* NICK IN USE */
        {
            char *nick = xmlnode_get_tag_data(jp->iq, "nickname");
            
            if (nick == NULL)
            { 
                /* no more nick's to try.. pity */
                jutil_error(jp->x, TERROR_CONFLICT);
		jp->aux1 = (void *) NULL;
		xhash_zap(ircdata->sessions, s->id);
		xhash_walk(s->channels, _irc_cleanup_channel, NULL);
		xhash_free(s->channels);
		if(s->buffer != NULL)
		{
		    free(s->buffer);
		    s->buffer = NULL;
		}
		pool_free(s->p);
		mio_close(m);
                pool_free(p);
                return;
            }
            else
            { /* try another nick */
                mio_write(m, NULL, spools(p, "NICK ", nick, "\n", p), -1);
                pool_free(p);
                return;
            }
        }
	/* RPL_WELCOME : we are logged in, can now join a channel */
        else if(j_strcmp(icpr->cmd, "001") == 0) 
        {
            char *ptr = strchr(icpr->next, ' ');
            if(ptr == NULL)
            { 
	        /* something's wrong.. we can't grab our nick */
	        log_warn(ZONE,"[irc_parse_phase_one] Unable to grab our nick !");
                mio_close(m);
                pool_free(p);
                return;
            }
            ptr[0] = '\0';
            /* note our nick */
            s->nick = pstrdup(s->p, icpr->next);
	    if (s->state == sess_ENTER)
	    {
		/* go into phase one */
		mio_reset(m, irc_parse_phase_one, arg);
		/* and send the join command */
		mio_write(m, NULL, spools(p, "JOIN ", irc_get_channel(p, jid_full(jp->to)), " ", xmlnode_get_tag_data(jp->iq, "query/password"), "\n", p), -1);
		pool_free(p);
		return;
	    }
	    else if (s->state == sess_CREATE)
	    {
		char *address, *chan_name;

		/* did user specify a room name ? */
		if ((address = xmlnode_get_tag_data(jp->iq, "address")) == NULL) 
		{
		    /* randomly set a channel name */
		    ircdata->rand_room++;
		    chan_name = pmalloc(s->p,strlen("#parisestmagique")+5);
		    sprintf(chan_name,"%s%d","#parisestmagique",ircdata->rand_room);
		    log_debug(ZONE,"[irc_parse_phase_zero] No room name, randomly setting : %s", chan_name);
		}
		else
		{
		    char *ptr;
		    /* remove the @ if it was specified */
		    ptr = strchr(address,'@');
		    if (ptr != NULL)
		      ptr[0] = '\0';
		    /* XXX : add a # before chan_name if not specified ! */
		    chan_name = pmalloc(s->p, strlen(address));
		    strcpy(chan_name,address);
		}

		/* temp store channel name in aux1 */
		s->aux1 = (void *) chan_name;
		/* check that room does not already exist
		 * before going into next phase */
		mio_write(m, NULL, spools(p, "MODE ", chan_name, "\n", p), -1);
	    }
        }
	/* RPL_CHANNELMODEIS */
	else if ((j_strcmp(icpr->cmd, "324") == 0) && (s->state == sess_CREATE))
	{
	    log_error(ZONE,"[irc_parse_phase_zero] Room already exists");
	    jutil_error(jp->x,TERROR_CONFLICT);
	    deliver(dpacket_new(jp->x),NULL);
	    return;
	}
	/* ERR_NOSUCHCHANNEL : that's was expected, let's go on */
	else if (j_strcmp(icpr->cmd, "403") == 0)
	{
	    char *channel, *passwd;

	    log_debug(ZONE,"[irc_parse_phase_zero] Channel does not exist, can create it !");

	    /* retrieve channel name */
	    channel = (char *) s->aux1;
	    /* go into phase one */
	    mio_reset(m, irc_parse_phase_one, arg);
	    /* and join the channel */
	    if ((passwd = xmlnode_get_tag_data(jp->iq, "password")) != NULL) {
		mio_write(m, NULL, spools(p, "JOIN ", channel, " ", passwd, "\n", p), -1);
	    }
	    else
	    {
		mio_write(m, NULL, spools(p, "JOIN ", channel, "\n", p), -1);
	    }
	}
    }

    pool_free(p);
}

/* handle responses to enter get requests */
void irc_check_iq_get(mio m, int flag, void *arg, char *buffer, int bufsz)
{
    /* session is on aux1 of the jpacket */
    jpacket jp = (jpacket)arg;
    irc_session s = (irc_session)jp->aux1;
    char *line, *nick;
    pool p;

    if(flag == MIO_ERROR) return;

    if(flag == MIO_CLOSED)
    {
        /* when aux1 is not cleared, we hit an error */
        if(jp->aux1 != NULL)
        {
            xmlnode_put_attrib(jp->x, "type", "error");
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "error"), "Socket Error to IRC Server", -1);
            xmlnode_put_attrib(xmlnode_get_tag(jp->x, "error"), "code", "502");
            deliver(dpacket_new(jp->x), NULL);
            pool_free(s->p);
        }
        else
        {
            /* if aux1 is cleared, then everything is going as planned */
            deliver(dpacket_new(jp->x), NULL);
        }
        return;
    }

    if(flag == MIO_NEW)
    {
        /* create a random nick, based on the address, %X, of the MIO object */
        nick = pmalloco(m->p, 15);
        snprintf(nick, 15, "temp_%X", m);
#ifdef IDX_IRC
	mio_write(m, NULL, spools(s->p, "PASS ", IDX_IRC_PASSWD, "\n", s->p), -1);
#endif
        /* log into the IRC server */
        mio_write(m, NULL, spools(s->p, "USER jabber 0 * :jabber user\n", s->p), -1);
        mio_write(m, NULL, spools(s->p, "NICK ", nick, "\n", s->p), -1);
        /* when the irc server responds, we will be back up top, in MIO_BUFFER */
        return;
    }

    /* the rest of this function is normal data from the IRC server */

    /* add this data onto the buffer in progress */
    irc_add_to_buffer(s, buffer, bufsz);

    /* create the working pool */
    p = pool_new();

    /* loop for each irc line.. */
    while((line = irc_get_next_line(s, p)) != NULL)
    {
        irc_cmd_parsing_result icpr = (irc_cmd_parsing_result) malloc(sizeof(_irc_cmd_parsing_result));

	if (irc_parse_line(line,icpr) == -1)
	    continue;

        /* check for commands we are interested in */
        if(j_strcmp(icpr->cmd, "PING") == 0)
        { 
            /* most irc servers ping you when you first log in, 
	     * as a capabilities check */
            mio_write(m, NULL, spools(p, "PONG ", icpr->next, "\n", p), -1);
        }
        else if(j_strcmp(icpr->cmd, "433") == 0) /* NICK IN USE */
        {
            /* create another random nick, based on the address of the mio object */
            nick = pmalloco(m->p, 15);
            snprintf(nick, 15, "temp_%X", m + 1);
            /* log into the IRC server */
            mio_write(m, NULL, spools(s->p, "USER ", nick, " foo foo jabber_user@jabber.org\n", s->p), -1);
            mio_write(m, NULL, spools(s->p, "NICK ", nick, "\n", s->p), -1);
            /* when the irc server responds, we will be back up top, in MIO_BUFFER */
            pool_free(p);
            return;
        }
	/* Successfull login - check for channel mod */
        else if(j_strcmp(icpr->cmd, "001") == 0)
        { 
	  if (s->state == sess_ENTER)
	  {
	    mio_write(m, NULL, spools(p, "MODE ", irc_get_channel(p, jid_full(jp->from)), "\n", p), -1);
	  }
        }
	/* MODE REPLY -- ERR_NOSUCHCHANNEL : Channel not found */
        else if(j_strcmp(icpr->cmd, "403") == 0) 
        {
	    /* if it's an enter request, send an error */
	    /* and free the session */
	    if (s->state == sess_ENTER) 
	    {
		jutil_iqresult(jp->x);
		jutil_error(jp->x, TERROR_NOTFOUND);
		pool_free(s->p);
		jp->aux1 = NULL;
		mio_close(m);
		pool_free(p);
		return;
	    } 
        }
	/* MODE REPLY -- RPL_CHANNELMODEIS */
        else if(j_strcmp(icpr->cmd, "324") == 0) 
        { 
            char *ptr = strchr(icpr->next, ' ');
            icpr->next = strchr(ptr, ' ');
            icpr->next++;
            
            /* we are looking for the 'k' mode.. (key) */
            if(strchr(icpr->next, 'k') != NULL)
                xmlnode_insert_tag(jp->iq, "password");

            /* keep the session and send the jpacket */
            jp->aux1 = NULL;
            mio_close(m);
            pool_free(p);
            return;
        }
            
    }

    /* free the working pool */
    pool_free(p);
}

void irc_check_browse_get(mio m, int flag, void *arg, char *buffer, int bufsz)
{
    jpacket jp = (jpacket)arg;
    irc_session s = (irc_session)jp->aux1;
    char *line, *nick;
    pool p;

    if (s == NULL) {
	log_error(ZONE,"[irc_check_browse_get] no session for user %s", jid_full(jp->from));
	return;
    }

    if(flag == MIO_ERROR) 
    {
        log_debug(ZONE, "[irc_check_browse_get] Error on browse socket");
        return;
    }

    if(flag == MIO_CLOSED)
    {
        /* return the packet, w/ or w/o the browse results */
        log_debug(ZONE, "[irc_check_browse_get] Closing browse request socket");
        deliver(dpacket_new(jp->x), NULL);
        if(s->buffer != NULL) 
        {
            free(s->buffer);
            s->buffer = NULL;
        }
        pool_free(s->p);
        return;
    }

    if(flag == MIO_NEW)
    {
        nick = pmalloco(m->p, 15);
        snprintf(nick, 15, "temp_%X", m);
        log_debug(ZONE, "[irc_check_browse_get] Logging with nick : %s", nick);
#ifdef IDX_IRC
	mio_write(m, NULL, spools(s->p, "PASS ", IDX_IRC_PASSWD, "\n", s->p), -1);
#endif
        mio_write(m, NULL, spools(s->p, "USER jabber 0 * :jabber foo\n", s->p), -1);
        mio_write(m, NULL, spools(s->p, "NICK ", nick, "\n", s->p), -1);
        return;
    }

    /* add the irc data to the buffer */
    irc_add_to_buffer(s, buffer, bufsz);

    /* create a new working pool */
    p = pool_new();

    while((line = irc_get_next_line(s, p)) != NULL)
    {
        irc_cmd_parsing_result icpr = (irc_cmd_parsing_result) malloc(sizeof(_irc_cmd_parsing_result));

	if (irc_parse_line(line,icpr) == -1)
	    continue;

        /* check commands */
        if(j_strcmp(icpr->cmd, "PING") == 0)
        {
            mio_write(m, NULL, spools(p, "PONG ", icpr->next, "\n", p), -1);
        }
        else if(j_strcmp(icpr->cmd, "433") == 0) /* NICK IN USE */
        {
            log_debug(ZONE, "[irc_check_browse_get] Nick in use.. crap");
            mio_close(m);
            pool_free(p);
            return;
        }
        else if(j_strcmp(icpr->cmd, "001") == 0)
        {
	    char *channel;
            log_debug(ZONE, "[irc_check_browse_get] Logged in successfull, checking server list for %s", jp->from->user);
	    /* handle requests sent to a specific user */
	    if (jp->from->resource != NULL)
	    {
		user u;
		chans chan;

		log_debug(ZONE,"[irc_check_browse_get] YEHA USER !");
		/* check that user is in the channel */
		channel = irc_get_channel(p,jid_full(jp->from));
		chan = xhash_get(s->channels, channel);
		if (chan == NULL) {
		    log_error(ZONE,"[irc_check_browse_get] Unable to find channel !");
		    /* XXX : return an error to user */
		    return;
		}
		u = xhash_get(chan->users, jid_full(jp->from));
		if (u == NULL) {
		    log_error(ZONE,"[irc_check_browse_get] user not found in channel %s", channel);
		    jutil_error(jp->x, TERROR_NOTFOUND);
		    jutil_tofrom(jp->x);
		    deliver(dpacket_new(jp->x),NULL);
		    return;
		}
		mio_write(m, NULL, spools(p, "WHOIS :", u->user, "\n", p), -1);
	    }
	    /* handle requests sent to #channel%server@irc and #channel@irc */
	    else if ((channel = irc_get_channel(p,jid_full(jp->from))) != NULL) {
		log_debug(ZONE,"[irc_check_browse_get] YEHA CHANNEL !");
		mio_write(m, NULL, spools(p, "LIST :", channel, "\n", p), -1);
	    }
	    /* handle requests sent to irc (default server) and %server@irc */
            else if ((jp->from->user == NULL) || (irc_get_server(p,jid_full(jp->from)) != NULL)) {
		log_debug(ZONE,"[irc_check_browse_get] YEAH SERVER !");
                mio_write(m, NULL, "LIST\n", -1);
	    }
        }
	/* Channel Listing -- RPL_LIST "<channel> <# visible> :<topic>" */
        else if(j_strcmp(icpr->cmd, "322") == 0) 
        {
            char *channel;
            char *users;
            char *ptr;
            xmlnode x;

            log_debug(ZONE, "[irc_check_browse_get] Got a channel listing: %s", icpr->next);
            channel = strchr(icpr->next, ' ');
            icpr->next = strchr(++channel, ' ');
            icpr->next[0] = '\0';
            users = ++icpr->next;
            icpr->next = strchr(users, ' ');
            icpr->next[0] = '\0';
            icpr->next += 2;
            ptr = strpbrk(icpr->next, "&<>\"\':");
            if(ptr != NULL) ptr[0] = '\0';
            icpr->next = str_to_UTF8(p, icpr->next);
            irc_strip_mirc_codes(icpr->next);

	    x = xmlnode_insert_tag(jp->iq, "room");
            xmlnode_put_attrib(x, "type", "irc");
            xmlnode_put_attrib(x, "jid", spools(p, channel, "@", ircdata->i->id, p));
            xmlnode_put_attrib(x, "name", spools(p, channel, " (", users, ") :", icpr->next, p));
        }
	/* End of List -- RPL_LISTEND */
        else if(j_strcmp(icpr->cmd, "323") == 0) 
        {
	    char *channel = irc_get_channel(p,jid_full(jp->from));

	    /* server request */
            if (channel  == NULL) 
            {
                log_debug(ZONE, "[irc_check_browse_get] browse request completed");
                mio_close(m);
                pool_free(p);
                return;
            }

            log_debug(ZONE, "[irc_check_browse_get] Checking users for channel: %s", channel);
            /* room request */
            mio_write(m, NULL, spools(p, "WHO :", channel, "\n", p), -1);
        }
	/* RPL_WHOREPLY */
        else if(j_strcmp(icpr->cmd, "352") == 0)
        {
            char *ptr;
            xmlnode x = xmlnode_insert_tag(jp->iq, "user");

            log_debug(ZONE, "[irc_check_browse_get] got user list for channel");
            /* 352 my_nick #channel user host server nick bla :name */
            ptr = strchr(icpr->next, '#'); /* channel */
            icpr->next = strchr(ptr, ' '); /* user */
            ptr = strchr(++icpr->next, ' '); /* host */
            icpr->next = strchr(++ptr, ' '); /* server */
            ptr = strchr(++icpr->next, ' '); /*nick*/
            icpr->next = strchr(++ptr, ' '); /* bla */
            if(icpr->next != NULL) 
                icpr->next[0] = '\0';
            else
                continue;
            
            xmlnode_put_attrib(x, "name", ptr);
            xmlnode_put_attrib(x, "jid", spools(p, jid_full(jp->from), "/", shahash(ptr), p));
        }
	/* RPL_ENDOFWHO -- End of WHO list */
	/* RPL_ENDOFWHOIS -- End of WHOIS list */
        else if ((j_strcmp(icpr->cmd, "315") == 0) || (j_strcmp(icpr->cmd,"318") == 0))
        { 
            log_debug(ZONE, "[irc_check_browse_get] Successfull browse.. sending data", ZONE);
            mio_close(m);
            pool_free(p);
            return;
        }
	/* RPL_WHOISUSER -- Reply to WHOIS query */
	/* 311 <my_nick> <nick> <user> <host> * :<real name> */
	/* XXX : handle other responses from server : 312,313,317,318 cmd */
	else if (j_strcmp(icpr->cmd,"311") == 0)
	{
            char *ptr,*nick,*real_name;
            xmlnode x = xmlnode_insert_tag(jp->iq, "user");

            log_debug(ZONE, "[irc_check_browse_get] got info for user in channel");
            ptr = strchr(icpr->next, ' '); /* nick */
	    nick = ptr;
	    icpr->next = strchr(++ptr, ' '); /* user */
	    icpr->next[0] = '\0';
            ptr = strchr(++icpr->next, ' '); /* host */
            icpr->next = strchr(++ptr, ':'); /* real name */
	    icpr->next++;
	    real_name = icpr->next;
            
            xmlnode_put_attrib(x, "name", nick);
            xmlnode_put_attrib(x, "jid", spools(p, jid_full(jp->from), "/", shahash(nick), p));
	    xmlnode_insert_cdata(xmlnode_insert_tag(x,"last"),icpr->next,-1);
	}
    }

    pool_free(p);
}
